from fs.enums import ResourceType
from fs.errors import DirectoryExpected, ResourceNotFound, FileExpected
from fs.info import Info

from .baseresource import BaseAnVILResource
from .google import GoogleAnVILFile, DRSAnVILFile
from .hypertext import HypertextAnVILFile

class BaseAnVILFolder(BaseAnVILResource):
    def __init__(self, name, last_modified=None):
        self.initialized = False
        if name[-1] != "/": # required since anvil supports names the same as their containing directories
            self.name = name + "/"
        else:
            self.name = name
        self.last_modified = last_modified
        self.children = {}

    def lazily_init(fn):
        def lazywrapper(*args, **kwargs):
            self = args[0]
            #DEBUGGERY#print(f"lazy init <{self.__class__.__name__} \"{self.name}\">.{fn.__name__}({args},{kwargs})? {self.initialized}")
            if not self.initialized:
                self.lazy_init()
                self.initialized = True
            return fn(*args, **kwargs)
        return lazywrapper

    def lazy_init(self):
        raise NotImplementedError(f"{self.__class__.__name__}.lazy_init method is abstract and must be specified")

    def __hash__(self):
        return hash((self.name, self.__class__.__name__, self.last_modified))

    def __eq__(self, other):
        return (self.name, self.last_modified) == (other.name, other.last_modified)

    # allow dictionary-style access, with possible objs as keys
    @lazily_init
    def __getitem__(self, key):
        return self.children[key]

    # allow for <item> in :
    def __iter__(self):
        return iter(self.children)

    @lazily_init
    def keys(self):
        def sorter(k):
            if k[-1] == "/":
                k = "0"+k
            return k
        return sorted([k for k in self.children.keys()], key=sorter)

    @lazily_init
    def get_object_from_path(self, path):
        if path == "/" or path == "":
            return self
        #internally, using google-style no initial slash
        if path[0] == "/":
            path = path[1:]
        # if path represents a folder:
        if path[-1] == "/":
            split = path[:-1].split("/")
            for i in range(len(split)):
                split[i] += "/"
        else:
            split = path.split("/")
            for i in range(len(split[:-1])):
                split[i] += "/"
        base_obj = self
        for component in split:
            base_obj = base_obj[component]
        return base_obj
    
    def __setitem__(self, key, val):
        self.children[key] = val

    def getinfo(self):
        _result = {
            "basic": {
                "name": self.name,
                "is_dir": True,
            },
            "details": {
                "type": ResourceType.directory,
                "modified": self.last_modified
            }
        }
        return Info(_result)

    def is_linkable_file(self, fname):

        protocol = fname.split("://")[0]
        filename = fname.split("/")[-1]
        allowed_protocols = {
            "gs": GoogleAnVILFile,
            "drs": DRSAnVILFile,
            # the below is better done downloading a tsv and using Galaxy's rule-based uploader, for now
            #"http": HypertextAnVILFile,
            #"https": HypertextAnVILFile
        }
        blocked_file_prefixes = [
            "data-explorer?"
        ]
        if protocol in allowed_protocols:
            for bfp in blocked_file_prefixes:
                if filename.startswith(bfp):
                    return None
            return allowed_protocols[protocol]
        else:
            return None


from io import BytesIO

from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder

class ReferenceDataFolder(BaseAnVILFolder):
    def __init__(self, name, refs):
        super().__init__(name)
        self.refs = refs
    
    def lazy_init(self):
        self.init_references(self.refs)

    def init_references(self, refs):
        for source in refs:
            # source, e.g. hg38
            source_folder = RefereneDataSubfolder(source+"/")
            self[source_folder.name] = source_folder
            # reftype, e.g. axiomPoly_resource_vcf
            for reftype in refs[source]:
                reftype_folder = RefereneDataSubfolder(reftype+"/")
                source_folder[reftype_folder.name] = reftype_folder
                contents = ReferenceDataFile.make_rdfs(refs[source][reftype])
                for c in contents:
                    reftype_folder[c.name] = c

class RefereneDataSubfolder(BaseAnVILFolder):
    def __init__(self, name):
        super().__init__(name)
        self.initialized = True
    
    def lazy_init(self):
        pass

class ReferenceDataFile(BaseAnVILFile):
    def __init__(self, blob):
        self.blob_handle = blob
        self.name = blob.name.split("/")[-1]
        self.last_modified = blob.updated
        self.size = blob.size
        self.is_dir = False

    @classmethod
    def make_rdfs(cls, objs):
        result = []
        for o in objs:
            result.append(cls(objs[o]))
        return result

    def get_bytes_handler(self):
        buffer = BytesIO()
        self.blob_handle.download_to_file(buffer)
        buffer.seek(0)
        return buffer


# def get_objects_info(anvilobj, objects):
#     print("getting objects...")
#     for o in objects:
#         print("{}:\n\t{}".format(o, anvilobj.getinfo(o).raw))

# def list_dirs(anvilobj, folders):
#     print("listing directories...")
#     for d in folders:
#         print("{}:\n\t{}".format(d, anvilobj.listdir(d)))

# def run_all(anvilobj, files, folders):
#     get_objects_info(anvilobj, files+folders)
#     list_dirs(anvilobj, folders)


import datetime
from google.auth import credentials
import json


class WorkloadIdentityCredentials(credentials.Scoped, credentials.Credentials):
    def __init__(self, scopes):
        super(WorkloadIdentityCredentials, self).__init__()
        #print(f"Init with scopes={scopes}")
        self._scopes = scopes

    def with_scopes(self, scopes):
        return WorkloadIdentityCredentials(scopes=scopes)

    @property
    def requires_scopes(self):
        return False

    def refresh(self, request):
        #print(f"Refresh with scopes={self._scopes}")
        url = 'http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token'
        if self._scopes:
            url += '?scopes=' + ','.join(self._scopes)
        response = request(url=url, method="GET", headers={
                           'Metadata-Flavor': 'Google'})
        if response.status == 200:
            response_json = json.loads(response.data)
        else:
            raise RuntimeError('bad status from metadata server')
        self.token = response_json['access_token']
        self.expiry = datetime.datetime.utcnow(
        ) + datetime.timedelta(seconds=response_json['expires_in'])


from time import sleep

from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder


class TableEntriesFile(BaseAnVILFile):
    def __init__(self, name, itemsdict):
        self.name = name
        self.last_modified = "" # #TODO determine if this is ok
        string = "\t".join(
                [f"{x}" for x in itemsdict]
            )
        for i in range(len(list(itemsdict.values())[0])):
            string += "\n" + "\t".join(
                [f"{itemsdict[x][i]}" for x in itemsdict]
            )
        self.buffstr = string
        self.size = len(string)
    
    def get_bytes_handler(self):
        return self.string_to_buffer(self.buffstr)


class RootTablesFolder(BaseAnVILFolder):
    def __init__(self, einfo, wsref):
        self.name = "Tables/"
        self.einfo = einfo
        self.wsref = wsref
        super().__init__(self.name)

    def lazy_init(self):
        for ename in self.einfo:
            attribs = self.einfo[ename]["attributeNames"]
            eid = self.einfo[ename]["idName"]
            tf = TableFolder(ename, eid, attribs, self.wsref)
            self[tf.name] = tf


# terra 'entities' represent tables
class TableFolder(BaseAnVILFolder):
    def __init__(self, etype, eid, attribs, wsref):
        self.name = etype + "/"
        super().__init__(self.name)
        self.type = etype
        self.eid = eid
        self.wsref = wsref
        self.attribs = attribs # column names
    
    def lazy_init(self):
        self.make_contents()

    def make_contents(self):
        base_table = {
            name: [] for name in self.attribs
        }
        linked_files = []
        file_links = {}
        # get remote info
        resp = self.get_entity_info()
        for entry in resp:
            e_attrs = entry["attributes"]
            if entry["entityType"] == "cohort" and "query" in e_attrs:
                linked_files.append(TableDataCohort(entry["name"],e_attrs["query"]))
            for attr in self.attribs:
                addendum = ""
                if attr in e_attrs:
                    val = e_attrs[attr]
                    if val is None:
                        val = ""
                    if type(val) == dict: # if theres more processing, e.g. not a string
                        # mere string attributes
                        if val["itemsType"] == "AttributeValue":
                            addendum = ",".join(val["items"])
                        # dict attributes referring to another entity
                        elif val["itemsType"] == "EntityReference": #TODO is this the best way to handle entity references?
                            addendum = ",".join(
                                [x["entityName"] for x in val["items"]]
                            )
                    else:
                        val = str(val) # enforce string
                        addendum = val
                        # check if its a linkable file
                        efiletype = self.is_linkable_file(val)
                        if efiletype is not None:
                            if efiletype not in file_links:
                                file_links[efiletype] = []
                            file_links[efiletype].append(val)
                base_table[attr].append(addendum)
        
        for method in file_links:
            try:
                fresh_files = method.factory(file_links[method])
                linked_files.extend(fresh_files)
            except Exception as e:
                print(f"AnVILFS ERROR: SKIPPING FILE, could not be resolved due to the following error:")
                print(e)
                continue
        # if there are links, make them externally available by entityname_filename.tsv
        linked_files.append(TableEntriesFile(self.type + "_contents.tsv", base_table))
        for f in linked_files:
            self[f.name] = f

    def get_entity_info(self):
        resp = self.fapi.get_entities(
            self.wsref.namespace.name, 
            self.wsref.name,
            self.type)
        if resp.status_code == 200:
            return resp.json()
        else:
            resp.raise_for_status()



class TableDataCohort(BaseAnVILFile):

    def __init__(self, name, query):
        #query = attribs["query"]
        self.name = name + "_query_results.tsv"
        self.size = 1 # some placeholder? cant lazy load AND init with total size of results
        self.query = query
        self.last_modified = "" # #TODO determine if this is ok

    def get_bytes_handler(self):
        job = self.gc_bigquery_client.query(self.query)
        r = job.result()
        schema_keys = [e.name for e in r.schema]
        string = '\t'.join(schema_keys)
        for row in r:
            string += "\n" + '\t'.join(row.values())
        return self.string_to_buffer(string)



from .testconfig import config
from .unit import run_all as run_all_unit_tests
# from .integration import run_all as run_all_integration_tests


# setup
if "anvil" not in locals():
    from ..anvilfs import AnVILFS
    from google.cloud import storage
    anvil = AnVILFS(config["namespace_name"], 
                    config["workspace_name"])
    bucket_name = anvil.workspace.bucket_name
    blobs = storage.Client().get_bucket(bucket_name).list_blobs()
    files = []
    folders = []
    bucketfolder_prefixes = ["Other Data", "Files"]
    for blob in blobs:
        split = blob.name.split("/")
        whole_split = bucketfolder_prefixes + split
        files.append("/".join(whole_split))
        for i in range(0,len(whole_split[:-1])):
            segment = whole_split[0:-1*(i+1)]
            reformed = "/".join(segment) + "/"
            if reformed not in folders:
                folders.append(reformed)
            else:
                break
    print("Contents of AnVIL workspace '{}'".format(config["workspace_name"]))
    print("    discovered files: \n{}".format(files))
    print("    discovered folders: \n{}\n".format(folders))

if __name__ == "__main__":
    unit_results, unit_failures = run_all_unit_tests(anvil, files, folders)
    # integration_results, integration_failures =  run_all_integration_tests(anvil, files, folders)


from io import BytesIO
from os.path import commonprefix

from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder
from .reference import ReferenceDataFile, ReferenceDataFolder
from .tables import RootTablesFolder
from .workloadidentitycredentials import WorkloadIdentityCredentials
from .workspacebucket import OtherDataFolder, WorkspaceBucket


class Workspace(BaseAnVILFolder):
    def __init__(self, namespace_reference,  workspace_name):
        self.namespace = namespace_reference
        # collect workspace attributes that inform structure
        resp = self.fetch_api_info(workspace_name)
        self.bucket_name = resp["workspace"]["bucketName"]
        self.attributes = resp["workspace"]["attributes"]
        try:
            super().__init__(workspace_name, resp["workspace"]["lastModified"])
        except KeyError as e:
            print("Error: Workspace fetch_api_info({}) fetch failed".format(workspace_name))
    
    def lazy_init(self):
        if self.initialized:
            print(f"{self.name} already initialized!")
            return
        # STUFF THAT CAN BE LAZILY LOADED
        # Tables folder
        table_baf = RootTablesFolder(self.fetch_entity_info(), self)
        self[table_baf.name] = table_baf
        # bucket folder
        bucket_baf = OtherDataFolder(self.attributes, self.bucket_name)
        self[bucket_baf.name] = bucket_baf
        # ref data folder
        refs = self.ref_extractor(self.attributes)
        ref_baf = ReferenceDataFolder("Reference Data/", refs)
        self[ref_baf.name] = ref_baf
        # populate workspace data
        # workspacedata = dict(self.attributes)
        # blocklist_prefixes = [
        #     "referenceData_",
        #     "description"
        # ]
        # for datum in self.attributes:
        #     for blocked in blocklist_prefixes:
        #         if datum.startswith(blocked):
        #             del workspacedata[datum]
        # if workspacedata:
        #     _wsd = WorkspaceData("WorkspaceData.tsv", workspacedata)
        #     bucket_baf[_wsd.name] = _wsd
        # _wsb = WorkspaceBucket(self.bucket_name)
        # bucket_baf[_wsb.name] = _wsb

    def ref_extractor(self, attribs):
        # structure:
        # { "source": {
        #      "reftype": {urlstr, blob} }}
        result = {}
        google_buckets = {}
        for ref in [r for r in attribs if r.startswith("referenceData_")]:
            val = attribs[ref]
            refsplit = ref.split("_", 2) 
            source = refsplit[1]
            reftype = refsplit[2]
            if source not in result:
                result[source] = {}
            if reftype not in result[source]:
                result[source][reftype] = {}
            root_result_obj = result[source][reftype]
            if isinstance(val, dict):
                val = val["items"]
            elif isinstance(val, str):
                val = [val]
            for v in val:
                parsed = self.url_parser(v)
                if not parsed:
                    continue
                root_result_obj[v] = None # blob placeholder
                if parsed["schema"] == "gs":
                    if parsed["bucket"] not in google_buckets:
                        google_buckets[parsed["bucket"]] = []
                    google_buckets[parsed["bucket"]].append(v)
                else:
                    raise Exception("Other schemas not yet implemented")
        # determine max shared prefix to limit results from api call
        url_to_blob = {}
        for bucket in google_buckets:
            gs_pfx = f"gs://{bucket}/"
            pfxs = [x[len(gs_pfx):] for x in google_buckets[bucket]]
            prefix = commonprefix(pfxs)
            uproj = self.gc_storage_client.project
            _bucket = self.gc_storage_client.bucket(bucket, user_project=uproj)
            blobs = self.gc_storage_client.list_blobs(_bucket, prefix=prefix)
            for blob in blobs:
                url = gs_pfx + blob.name
                url_to_blob[url] = blob
        # go back and add blobs to result
        for source in result:
            src_vals = result[source]
            for reftype in src_vals:
                refs = src_vals[reftype]
                for url in refs:
                    refs[url] = url_to_blob[url]
        return result

    def url_parser(self, url):
        split = url.split("://", 1)
        # if this is not a url, ignore
        if len(split) == 1:
            return None
        schema = split[0]
        path =  split[1]
        components = path.split("/", 1)
        subcomponents = components[1].split("/")
        return {
            "schema": schema,
            "bucket": components[0],
            "path": components[1],
            "source": subcomponents[0],
            "filename": subcomponents[-1]
        }

    def fetch_api_info(self, workspace_name):
        fields = "workspace.attributes,workspace.bucketName,workspace.lastModified"
        resp = self.fapi.get_workspace(namespace=self.namespace.name, workspace=workspace_name, fields=fields)
        if resp.status_code == 200:
            return resp.json()
        else:
            resp.raise_for_status()

    def fetch_entity_info(self):
        resp = self.fapi.list_entity_types(namespace=self.namespace.name, workspace=self.name)
        if resp.status_code != 200:
            resp.raise_for_status()
        return resp.json()



import requests
import json
from io import BytesIO
import concurrent.futures

from google.auth import credentials
from google.cloud import storage

import gs_chunked_io as gscio

from .basefile import BaseAnVILFile
from .clientrepository import ClientRepository

class GoogleAnVILFile(BaseAnVILFile):
    def __init__(self, input):
        if type(input) == str and input.startswith("gs://"):
            #normal
            url = input
            _split = url[len("gs://"):].split("/")
            self.name = _split[-1]
            filepath = "/".join(_split[1:])
            self.blob = self.info_to_blob(_split[0], filepath)
            self.blob.reload()
            self.size = self.blob.size
            self.last_modified = self.blob.updated
            # blobs = self.gc_storage_client.list_blobs(_bucket, prefix=filepath)
            # #buck = self.client.get_bucket(_split[0])
            # self.blob = None
            # self.size = None
            # self.last_modified = None
            # for b in blobs:
            #     if b.name == filepath:
            #         self.blob = b
            #         self.size = b.size
            #         self.last_modified = b.updated
            #         break
            # if not self.blob and not self.size and not self.last_modified:
            #     raise Exception(f"blob '{self.name}' not found...")
        elif type(input) == dict:
            self.name = input["name"]
            self.blob = self.info_to_blob(input["bucket"], input["path"])
            self.size = input["size"]
            self.last_modified = input["last_modified"]
        #self.blob = buck.get_blob("/".join(_split[1:]))
    
    @classmethod
    def factory(cls, gslist):
        results = []
        for item in gslist:
            results.append(GoogleAnVILFile(item))
        return results

    def info_to_blob(self, source_bucket, path):
        # requires project, bucket_name, prefix
        kb = 1024
        mb = 1024*kb
        chunk_size = 200*mb
        uproj = self.gc_storage_client.project
        bucket = self.gc_storage_client.bucket(source_bucket, user_project=uproj)
        return storage.blob.Blob(path, bucket)#, chunk_size = chunk_size)

    def get_bytes_handler(self):
        #buff = BytesIO()
        #self.blob.download_to_file(buff)
        #buff.seek(0)
        return gscio.Reader(self.blob)

class DRSAnVILFile(GoogleAnVILFile):
    api_url = "https://us-central1-broad-dsde-prod.cloudfunctions.net/martha_v3"

    def __init__(self, input):
        token = ClientRepository().get_fapi_token()
        #api_prefix = "https://dataguids.org/ga4gh/dos/v1/dataobjects/" <- old news
        api_url = self.api_url
        #apistring = api_prefix + drsurl[len("drs://"):]
        response = requests.post(
            api_url,
            data = {
                "url": input
            },
            headers = {
                "Authorization": f"Bearer {token}"
            }
        )
        result = json.loads(response.text)
        gurl = result["gsUri"]
        super().__init__(gurl)

    @classmethod
    def factory(cls, drslist):
        # subfunction for threads
        def _load_url(drsuri, timeout):
            token = ClientRepository().get_fapi_token()
            url = cls.api_url
            r = requests.post(
                    url,
                    data = {
                    "url":drsuri
                    },
                    headers = {
                        "Authorization": f"Bearer {token}"
                    }
                )
            return json.loads(r.text)
        # thread pool maker
        def _pooler(inlist, maxworks=50):
            timeout = 60#seconds
            good_data = []
            bad_uris = []
            with concurrent.futures.ThreadPoolExecutor(max_workers=maxworks) as executor:
                # Start the load operations and mark each future with its URL
                future_to_url = {executor.submit(_load_url, url, timeout): url for url in inlist}
                for future in concurrent.futures.as_completed(future_to_url):
                    url = future_to_url[future]
                    try:
                        data = future.result()
                        if "gsUri" not in data:
                            print(f"DRS resolution error - received:\n\t{data}")
                            bad_uris.append(url)
                        else:
                            good_data.append(data)
                    except Exception as exc:
                        print('%r generated an exception: %s' % (url, exc))
                        raise exc
                        bad_uris.append(url)
                    else:
                        pass
            return good_data, bad_uris

        # first pass
        file_objects = []
        good, bad = _pooler(drslist)
        # retry
        good_retries, bad_retries = _pooler(bad)
        if bad_retries:
            print(f"Unable to resolve the following URIs:\n{bad_retries}")
        total_goods = good + good_retries
        gs_info = [
            {
                "gsUri": x["gsUri"],
                "bucket": x["bucket"],
                "path": x["name"],
                "size": x["size"],
                "name": x["fileName"],
                "last_modified": x["timeUpdated"]
            } 
            for x in total_goods]
        # make google bucket objects
        for item in gs_info:
            file_objects.append(GoogleAnVILFile(item))
        return file_objects
